import torch
import alive_progress
from .modelbase import ModelBase
class Trainer(ModelBase):
    """
    A class for training, evaluating and testing PyTorch models.

    Parameters
    ----------
        model (torch.nn.Module): The PyTorch model to be trained, evaluated and tested.
        optimizer (torch.optim.Optimizer): The optimizer to be used for training.
        loss_fn (torch.nn.Module): The loss function to be used for training and evaluation.
        early_stop (object): The early stopping criteria to stop training early if validation loss stops improving.
        scheduler (torch.optim.lr_scheduler._LRScheduler): The learning rate scheduler to be used for training.
        train_loader (torch.utils.data.DataLoader): The DataLoader for the training dataset.
        eval_loader (torch.utils.data.DataLoader): The DataLoader for the evaluation dataset.
        test_loader (torch.utils.data.DataLoader): The DataLoader for the testing dataset.
        device (torch.device): The device to be used for training, evaluation and testing.
        verbose (bool): If True, prints the training and evaluation progress. Default is True.
        train_verbose (bool): If True, prints the training progress. Default is True.
        train_verbose_step (int): The number of steps to print during the training progress. Default is 5.
        eval_verbose (bool): If True, prints the evaluation progress. Default is False.
        n_classes (int): The number of classes for the classification task. Default is 10.



    Methods
    ----------
        epoch (int): The current epoch number.
        epochs (int): The total number of epochs to train the model for.
        train_loss (float): The training loss.
        eval_loss (float): The evaluation loss.
        n_classes (int): The number of classes for the classification task.
        evaluation_target (numpy.ndarray): The target values for the evaluation dataset.
        evaluation_prediction (numpy.ndarray): The predicted values for the evaluation dataset.

    Examples
    --------
    >>> model = Model()
    >>> optimizer = optim.Adam(model.parameters(), lr=0.01)
    >>> loss_fn = nn.CrossEntropyLoss()
    >>> device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    >>> early_stop = EarlyStopper(patience=15, min_delta=0.01)
    >>> scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, 'min', factor=0.1, patience=5, verbose=True)
    >>> trainer = Trainer(model=model, optimizer=optimizer, loss_fn=loss_fn, early_stop=early_stop, scheduler=scheduler,
    ...                   train_loader=train_loader, eval_loader=eval_loader, test_loader=test_loader, device=device,
    ...                   verbose=True, eval_verbose=True, n_classes=9)
    >>> trainer.run()
    >>> trainer.evaluate_result()
    >>> trainer.save_model("model.pth")
    """
    def __init__(self, model, optimizer, loss_fn, early_stop=None, scheduler=None,
                 train_loader=None, eval_loader=None, test_loader=None,
                 device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu'),
                 verbose: bool=True, train_verbose: bool=True, train_verbose_step: int=5, eval_verbose: bool=False, n_classes=10):
        if not train_loader and not eval_loader:
            print("Warning: No training and validation dataloader is provided to the Trainer()")

        self.model = model.to(device)
        self.train_loader = train_loader
        self.eval_loader = eval_loader
        self.test_loader = test_loader
        self.optimizer = optimizer
        self.loss_fn = loss_fn
        self.early_stop = early_stop
        self.scheduler = scheduler
        self.device = device
        self.verbose = verbose
        self.train_verbose = train_verbose
        self.train_verbose_step = train_verbose_step
        self.eval_verbose = eval_verbose

        # Training Parameters
        self.epoch = 0
        self.epochs = 0
        self.train_loss = 0

        # Evaluation Parameters
        self.eval_loss = None
        self.n_classes = n_classes
        self.evaluation_target = None
        self.evaluation_prediction = None

    def run(self, epochs: int = 10):
        self.epochs = epochs
        for self.epoch in range(self.epochs):
            if self.train_verbose and self.verbose:
                print(f"Epoch {self.epoch + 1} of {self.model.__class__.__name__}\n-------------------------------")
            self.train()
            if self.eval_loader:
                self.eval()
                if self.early_stop and self.early_stop.early_stop(validation_loss=self.eval_loss, verbose=self.eval_verbose and self.verbose):
                    break
        self.test()


    def train(self):
        """
        Trains the model on the training dataset.
        """
        self.model.train()
        for batch, (feature, target) in enumerate(self.train_loader):
            feature, target = feature.to(self.device), target.to(self.device)
            prediction = self.model(feature)
            try:
                self.train_loss = self.loss_fn(prediction, target)
            except:
                target = target.squeeze(1)
                self.train_loss = self.loss_fn(prediction, target)
                # self.train_loss = self.loss_fn(prediction.argmax(1).float(), target.float())
            self.optimizer.zero_grad()
            self.train_loss.backward()
            self.optimizer.step()
            if self.train_verbose and (batch + 1) % int(len(self.train_loader)/self.train_verbose_step) == 0 and self.verbose:
                print(f"Epoch [{self.epoch + 1}/{self.epochs}], Step [{batch + 1}/{len(self.train_loader)}], "
                      f"Loss: {self.train_loss.item():.4f}, Learning Rate {self.optimizer.param_groups[0]['lr']}")

    def eval(self):
        """
        Evaluates the model on the evaluation dataset.
        """
        with torch.no_grad():
            self.model.eval()
            for batch, (feature, target) in enumerate(self.eval_loader):
                feature, target = feature.to(self.device), target.to(self.device)
                prediction = self.model(feature)
                try:
                    self.eval_loss = self.loss_fn(prediction, target)
                except:
                    target = target.squeeze(1)
                    self.eval_loss = self.loss_fn(prediction, target)
            self.scheduler.step(self.eval_loss)

    def test(self):
        """
        Tests the model on the testing dataset.
        """
        import torchmetrics
        self.model.eval()
        self.test_target = []
        self.test_prediction = []
        with torch.no_grad():
            self.model.eval()
            for batch, (feature, target) in enumerate(self.test_loader):
                feature, target = feature.to(self.device), target.to(self.device)
                prediction = self.model(feature)
                self.test_target.append(target)
                self.test_prediction.append(prediction)
            self.test_target = torch.stack(self.test_target, dim=0).reshape(-1)
            self.test_prediction = torch.stack(self.test_prediction, dim=0).reshape(-1, self.n_classes)
        #from sklearn.metrics import classification_report, confusion_matrix, roc_curve, accuracy_score
        #self.cm = confusion_matrix(test_target=self.target, y_pred=self.fused_prediction)

        self.evaluation_target = self.test_target.detach().cpu().numpy()
        self.evaluation_prediction = self.test_prediction.argmax(1).detach().cpu().numpy()

    def save_model(self, path="model.pth"):
        """
        Saves the PyTorch model to the specified path.
        """
        torch.save(self.model, path)

    def __str__(self):
        """
        Returns a string representation of the model, optimizer, loss function and device used for training.
        """
        model_name = self.model.__class__.__name__
        optimizer_name = self.optimizer.__class__.__name__
        loss_fn_name = self.loss_fn.__class__.__name__
        device_name = self.device
        return f"Name of Model:\t\t{model_name}\n" \
               f"Used Optimizer:\t\t{optimizer_name}\n" \
               f"Used Loss Function:\t{loss_fn_name}\n" \
               f"Used Device:\t\t{device_name}"



from .trainer import Trainer
from .utilities import EarlyStopper
from .fuser import ModelFuser
from .user import User


__all__ = ["Trainer",
           "EarlyStopper",
           "ModelFuser",
           "User"]



import torch
from .modelbase import ModelBase
class User(ModelBase):
    """
    A class for evaluating and using a model.

    Parameters
    ----------
    model: torch.nn.Module
        The user model to evaluate.
    n_classes: int
        The number of classes in the dataset.
    dataset: torch.utils.data.DataLoader
        The test dataset to evaluate the model on.
    device: torch.device, optional
        The device to use for evaluation (default: 'cuda:0' if available, otherwise 'cpu').

    Examples
    --------
    >>> model = torch.load("model_medmnist/fifth_model.mdl")
    >>> use_model = User(model, 9, test_loader, device)
    >>> use_model.evaluate()
    """
    def __init__(self, model, n_classes, dataset, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
        self.model = model.to(device)
        self.n_classes = n_classes
        self.test_loader = dataset
        self.device = device
    def evaluate(self):
        """
        Evaluate the user model on the test dataset and calculate evaluation metrics.
        """
        self.model.eval()
        self.test_target = []
        self.test_prediction = []
        with torch.no_grad():
            self.model.eval()
            for batch, (feature, target) in enumerate(self.test_loader):
                feature, target = feature.to(self.device), target.to(self.device)
                prediction = self.model(feature)
                self.test_target.append(target)
                self.test_prediction.append(prediction)
            self.test_target = torch.stack(self.test_target, dim=0).reshape(-1)
            self.test_prediction = torch.stack(self.test_prediction, dim=0).reshape(-1, self.n_classes)
        #from sklearn.metrics import classification_report, confusion_matrix, roc_curve, accuracy_score
        #self.cm = confusion_matrix(test_target=self.target, y_pred=self.fused_prediction)

        self.evaluation_target = self.test_target.detach().cpu().numpy()
        self.evaluation_prediction = self.test_prediction.argmax(1).detach().cpu().numpy()
        self.evaluate_result()



class ModelBase:
    """
    A base class for machine learning models.

    Attributes:
        cm (numpy.ndarray): Confusion matrix computed during evaluation.
        evaluation_target (numpy.ndarray): True labels for evaluation.
        evaluation_prediction (numpy.ndarray): Predicted labels for evaluation.

    """
    def evaluate_result(self):
        """
        Computes and displays evaluation metrics such as accuracy, precision, recall, f1-score, and confusion matrix.
        """
        import numpy as np
        import matplotlib.pyplot as plt
        from sklearn.metrics import classification_report, confusion_matrix, roc_curve, accuracy_score
        target = self.evaluation_target
        prediction = self.evaluation_prediction
        self.cm = confusion_matrix(y_true=target, y_pred=prediction)
        import numpy as np
        import pandas as pd
        import copy
        cm = self.cm
        FP = cm.sum(axis=0) - np.diag(cm)
        FN = cm.sum(axis=1) - np.diag(cm)
        TP = np.diag(cm)
        TN = cm.sum() - (FP + FN + TP)

        total_cm = copy.deepcopy(cm)
        np.fill_diagonal(total_cm, 0)
        total_wrong = total_cm.sum()

        # Sensitivity, hit rate, recall, or true positive rate
        TPR = TP / (TP + FN)
        # Specificity or true negative rate
        TNR = TN / (TN + FP)
        # Precision or positive predictive value
        PPV = TP / (TP + FP)
        # Negative predictive value
        NPV = TN / (TN + FN)
        # Fall out or false positive rate
        FPR = FP / (FP + TN)
        # False negative rate
        FNR = FN / (TP + FN)
        # False discovery rate
        FDR = FP / (TP + FP)

        F1_Score = 2 * (PPV * TPR) / (PPV + TPR)
        # Overall accuracy
        ACC = TP / (TP + FP + FN)
        unique, counts = np.unique(target, return_counts=True)
        Support = counts
        weighted_f1_score = np.average(F1_Score, weights=Support)
        avg_ACC = sum(TP) / (sum(TP) + sum(FP))
        data = {"Accuarcy (Micro average f1-score)": avg_ACC, "Precision": PPV, "Recall": TPR, "F1 Score": F1_Score, "weighted f1-score": weighted_f1_score, "Support": Support}
        pd.DataFrame(data)
        print(cm)
        #print(f"Support: {Support}")
        #print(f"Accuarcy (Micro average f1-score): {np.around(avg_ACC, decimals=4)}")
        #print(f"weighted f1-score {np.around(weighted_f1_score, decimals=4)}")
        #print(f"class Accuarcy: {np.around(ACC, decimals=4)}")
        #print(f"class Precision: {np.around(PPV, decimals=4)}")
        #print(f"class Recall: {np.around(TPR, decimals=4)}")
        #print(f"class F1_Score: {np.around(F1_Score, decimals=4)}")
        print(f"Total Wrong: {total_wrong}")


        from sklearn.metrics import classification_report, confusion_matrix, roc_curve
        print(classification_report(target, prediction))

    def plot_evaluate_result(self):
        """
        Plots the confusion matrix using seaborn heatmap.
        """
        import seaborn as sns
        import pandas as pd
        import matplotlib.pyplot as plt
        df_cm = pd.DataFrame(self.cm)
        sns.heatmap(df_cm, annot=True)
        plt.show()






import numpy as np
class EarlyStopper:
    """
    Early stopper that stops training if the validation loss does not improve
    for a given number of epochs.

    Parameters
    ----------
        patience (int): The number of epochs to wait if the validation loss
            does not improve before stopping early.
        min_delta (float): The minimum change in validation loss to be
            considered as an improvement.

    Attributes:
        patience (int): The number of epochs to wait if the validation loss
            does not improve before stopping early.
        min_delta (float): The minimum change in validation loss to be
            considered as an improvement.
        counter (int): The number of epochs that the validation loss has not
            improved.
        min_loss (float): The best validation loss so far.
    """
    def __init__(self, patience=1, min_delta=0):
        self.patience = patience
        self.min_delta = min_delta
        self.counter = 0
        self.min_loss = np.inf

    def early_stop(self, validation_loss, verbose=False):
        if validation_loss < self.min_loss - self.min_delta:
            self.counter = 0
            self.min_loss = validation_loss
            if verbose:
                print(f"Early Stop: Counter reset. Min loss is {self.min_loss}.")
        else:
            self.counter += 1
            if verbose:
                print(f"Early Stop: Counter increased to {self.counter}. Min loss is {self.min_loss}. Patience is set to {self.patience}.")
            if self.counter >= self.patience:
                if verbose:
                    print("Early Stop: Patience reached. Stopping early.")
                return True
        return False



import torch
from .modelbase import ModelBase
from torch.utils.data import DataLoader
class ModelFuser(ModelBase):
    """
    Class for fusing the predictions of multiple models.

    Parameters
    ----------
    *models : ModelBase objects
        Multiple instances of ModelBase objects to be fused.
    test_loader : torch.utils.data.DataLoader
        DataLoader object for test dataset.
    device : torch.device, optional
        Device to use for computation. Default is 'cuda:0' if
        torch.cuda.is_available() is True, otherwise 'cpu'.

    Attributes
    ----------
    models : list
        List of ModelBase objects.
    n_models : int
        Number of models to fuse.
    test_loader : torch.utils.data.DataLoader
        DataLoader object for test dataset.
    n_classes : int
        Number of classes.
    device : torch.device
        Device used for computation.
    target : numpy.ndarray
        Target values of test dataset.
    fused_prediction : numpy.ndarray
        Fused predictions of all models for test dataset.
    evaluation_target : numpy.ndarray
        Target values used for evaluation.
    evaluation_prediction : numpy.ndarray
        Fused predictions used for evaluation.
    probability : numpy.ndarray
        Fused probabilities for all classes.



    Examples
    --------
    >>> model_a = ModelA()
    >>> model_b = ModelB()
    >>> dataset = MyDataset('test')
    >>> loader = DataLoader(dataset, batch_size=32)
    >>> fuser = ModelFuser(model_a, model_b, test_loader=loader)
    >>> fuser.fuse()
    >>> fuser.evaluate_result()
    >>> fuser.plot_evaluate_result()

    """
    def __init__(self, *models, test_loader=None, device=torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')):
        """
        Initializes the ModelFuser object.

        Parameters
        ----------
        *models : ModelBase objects
            Multiple instances of ModelBase objects to be fused.
        test_loader : torch.utils.data.DataLoader
            DataLoader object for test dataset.
        device : torch.device, optional
            Device to use for computation. Default is 'cuda:0' if
            torch.cuda.is_available() is True, otherwise 'cpu'.
        """

        self.models = models
        self.n_models = len(models)
        self.test_loader = test_loader
        self.n_classes = 9
        self.device = device
        assert isinstance(test_loader, torch.utils.data.DataLoader), "Loader is not of type DataLoader"
    def fuse(self):
        """
        Fuses the predictions of all models.
        """
        import torch
        import numpy as np
        import pandas as pd
        assert self.test_loader, "Kein Testdatensatz!"

        predictions = []
        targets = []
        for model in self.models:
            prediction = []
            model.to(self.device)
            model.eval()
            predictions_per_model = []
            targets_per_model = []
            with torch.no_grad():
                for batch, (feature, target) in enumerate(self.test_loader):
                    feature, target = feature.to(self.device), target.to(self.device)
                    prediction = model(feature)
                    targets_per_model.append(target)
                    predictions_per_model.append(prediction)
            predictions.append(torch.stack(predictions_per_model, dim=0))
            targets.append(torch.stack(targets_per_model, dim=0))
        predictions = torch.stack(predictions, dim=0)
        target = torch.stack(targets, dim=0)
        probabilities = torch.nn.functional.softmax(predictions, dim=3).detach().cpu().numpy()
        combined_probability = np.apply_along_axis(lambda row: self.combine_probabilities(row), axis=1, arr=probabilities.reshape([-1, self.n_models * self.n_classes]))
        all_pred = predictions.detach().cpu().numpy().argmax(3).reshape((self.n_models, -1))
        all_target = target[0].detach().cpu().numpy().reshape(-1)
        data = np.vstack([all_target, all_pred]).T
        probabilities = probabilities.reshape((self.n_models, -1, self.n_classes))
        for i in probabilities:
            data = np.hstack([data, i])
        # get_class(all_pred)
        fused_prediction = np.apply_along_axis(lambda row: self.apply_voting(row), axis=1, arr=data)
        result = np.vstack([all_target, fused_prediction])
        self.target = all_target.reshape(-1)
        self.fused_prediction = fused_prediction.reshape(-1)
        self.evaluation_target = self.target
        self.evaluation_prediction = self.fused_prediction
        self.probability = combined_probability
        # df = pd.DataFrame(data=data, columns=["Model1", "Model2", "Model3", "Target"])
        # df["Probabilites"] = probabilities[0]
        # print("HERE")
        from sklearn.metrics import confusion_matrix
        # confusion_matrix(predictions.detach().cpu().numpy(), self.y_pred.argmax(1).detach().cpu().numpy())
        # return result

    def combine_probabilities(self, row):
        """
        Computes the combined probability of all models for each class.

        Parameters
        ----------
        row : numpy.ndarray
            Row of fused probabilities for all models.

        Returns
        -------
        numpy.ndarray
            Combined probability for each class.
        """
        import numpy as np
        from scipy.special import softmax
        probabilities = row.reshape([self.n_models, self.n_classes])
        #probabilities = softmax(probabilities, axis=1)
        combined_probability = np.mean(probabilities, axis=0)
        return combined_probability

    def apply_voting(self, row):
        """
        Applies voting mechanism to select final prediction.

        Parameters
        ----------
        row : numpy.ndarray
            Row of fused predictions and probabilities for all models.

        Returns
        -------
        int
            Final prediction selected by voting mechanism.
        """
        import numpy as np
        import math
        from scipy.special import softmax
        vote_array = row[1:self.n_models+1].astype(int)
        total_votes = self.n_models
        votes_for_results = math.ceil(total_votes/2)
        max_val = np.argmax(np.bincount(vote_array))
        unique_vals, counts = np.unique(vote_array, return_counts=True)

        if counts.shape[0] <= votes_for_results:
            # print("Voting Successful")
            voted_value = max_val
            return voted_value
        else:
            range_probabilities = self.n_classes * self.n_models
            probabilities = row[-range_probabilities:].reshape([self.n_models, self.n_classes])
            probabilities = softmax(probabilities, axis=1)
            combined_probability = np.mean(probabilities, axis=0)
            voted_value = np.argmax(combined_probability)
            return voted_value



from setuptools import setup, find_packages
import codecs
import os

VERSION = '0.0.1'
DESCRIPTION = 'Simplifying working with PyTorch'
LONG_DESCRIPTION = 'TorchEase is a Python package that simplifies working with PyTorch, a popular deep learning framework. It provides a set of easy-to-use tools for training, testing, and deploying machine learning models. TorchEase includes features like early stopping, model fusing, and model evaluation, among others, which help streamline the model development process. Additionally, TorchEase is designed to be highly modular, making it easy to integrate into existing PyTorch projects.'

# Setting up
setup(
    name="TorchEase",
    version=VERSION,
    author="Philipp Steigerwald",
    author_email="info@b-stream.info",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=[],
    keywords=['python', 'torch', 'pytorch', 'early stop', 'trainer', 'tensor'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)